/* routines internal to fishnet's amphibian module. */
/* students are expected to provide connect, close, write, and
   bind functions */

#ifndef _AMPHIBIAN_H
#define _AMPHIBIAN_H
#include <sys/socket.h>
#include "amphibian_app.h"

/* opaque data type that acts as a handle for the transport
   protocol's communication.  It's roughly like a file
   descriptor or file pointer, but we can just pass pointers
   around instead of integers and it's ok. */
struct transport_state;

/* how the app tells the transport protocol to pass
   events, and stores a small amount of state */
/* represents the passing of information up from the
   transport protocol to the application */
struct transport_app_state {
  /* receive_data will take a buffer and deliver some part
     of it to the application.  receive_data returns the
     number of bytes read, possibly 0 */
  int (*receive_data)(struct transport_app_state *app, 
                      const char *data, 
                      int content_length);

  /* called *after* all data in the connection has been
     received, received_fin tells the application 
     that no more data will be received. */
  void (*received_fin)(struct transport_app_state *ts);

  /* called whenever the other end has acknowledged data,
     indicates that perhaps more can be written into a
     buffer, or that a connection is now established */
  void (*ready_for_data)(struct transport_app_state *ts);

  /* called when the transport connection dies, which may be
     unexpectedly if the other side crashes or is
     unresponsive */
  void (*destroy)(struct transport_app_state *app);

  /* for the app-only use */
  int fd;  /* a file descriptor for the app to use */
  void *extra; /* a pointer for the app to use. */

  /* for the transport protocol-only use */
  struct transport_state *ts; /* may be a pointer to itself, or be unused */
};

/* for fishnet code... when initializing the amphibian
   module (and thus becoming ready to service fishapps),
   fill in this structure, and hand it to init_amphibian.
   DO NOT FREE THE STRUCTURE OR ALLOCATE IT ON THE STACK. */

/* represents the passing of information down from the
   application to the transport protocol. */
struct amphibian_protocol {
  /* a non-blocking connect - package up the SYN packet and send it.
     the application will get notified that the connect is complete
     because the 'ready_for_data' callback will be called */
  struct transport_app_state *(*connect)(int remoteaddr, int remoteport);
  /* a non-blocking close - no more data will be sent on this socket,
     no more data should be received on it either. */
  void (*close)(struct transport_app_state *ts); 
  /* append data to the send queue.  it is the caller's responsibility
     to make sure there is enough buffer space using avail_write_buffer */
  void (*write)(struct transport_app_state *ts, const char *buf, int bytes); 
  /* bind to a local port.  When a syn is received on that port, call 
     acceptfn with a new transport_app_state structure, the address and
     port that sent the SYN, and a user_data parameter passed in the call
     to bind() */
  void (*bind)(int port, void (*acceptfn)(struct transport_app_state *ts, 
                                          int remote_address,
                                          int remote_port,
                                          void *user_data),
               void *user_data);
  /* how much space is there in the write buffer?  used
     before writing data.  should return
     TRANSPORT_PAYLOAD_SIZE or zero if a stop and wait
     sender, more if its a sliding window transport
     protocol */
  int (*avail_write_buffer)(struct transport_app_state *ts);
};

/* called to initialize the amphibian module, pass in a filled 
   out structure from above, and the fish address you're using. */
int amphibian_init(struct amphibian_protocol *protocol, int my_fish_address);
#endif
