/* the routines that manufacture the illusion that the
   socket-like function calls talk across a user-kernel
   boundary when, in fact, they connect across a process
   boundary to the fishnet node. */

#include <unistd.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <assert.h>
#include "amphibian_app.h"

/* returns the new socket */
/* domain should be AF_FISH, type SOCK_STREAM, protocol is unused */
int fishsockets_socket(int domain, int type, int protocol) {
  if(domain == AF_FISH) {
    struct sockaddr_un saun;
    int ret = socket(AF_UNIX, SOCK_STREAM, 0);
    const char *address;

    memset(&saun, 0, sizeof(struct sockaddr_un));
    saun.sun_family = AF_UNIX;
    saun.sun_path[0]='\0';
    if((address = getenv("AMPHIBIAN")) != NULL) {
      sprintf(saun.sun_path+1, "fish-%s-%s", getenv("USER"), address);
      if( connect(ret, (const struct sockaddr *)&saun, sizeof(struct sockaddr_un)) < 0) {
        fprintf(stderr, "amphibian: unable to connect fishsocket: %s\n", strerror(errno));
        fprintf(stderr, "amphibian: is your solution running with address %s?\n", address);
        close(ret);
        return -1;
      }
    } else {
      int i;
      ret = -1;
      for(i=1; i<=100 && ret == -1; i++) {
        sprintf(saun.sun_path+1, "fish-%s-%d", getenv("USER"), i);
        ret = connect(ret, (const struct sockaddr *)&saun, sizeof(struct sockaddr_un));
      }
      fprintf(stderr, "amphibian_app: unable to find a working fishsocket: %s\n", strerror(errno));
    }
    return ret;
  } else {
    return socket(domain, type, protocol);
  }
}

/* returns 0 on success, -1 on failure */
int fishsockets_connect(int sockfd, const struct sockaddr *serv_addr, socklen_t addrlen) {
  if(serv_addr != NULL && addrlen != 0) {
    if( ((const struct sockaddr_fish *)serv_addr)-> sfish_family == AF_FISH ) {
      struct connection_request req;
      char status_buf[20];
      int readbytes;
      assert( addrlen == sizeof(struct sockaddr_fish));
      req.connect_not_bind = TRUE;
      memcpy(&req.dst, serv_addr, addrlen);
      if(write(sockfd, &req, sizeof(req)) < (int)sizeof(req)) {
        return -1;
      }
      if((readbytes = read(sockfd, &status_buf, 20)) <= 0) {
        return -1;
      }
      if(strncmp(CONNECT_FAILED, status_buf, 
                 strlen(CONNECT_FAILED)) == 0) {
        errno = ETIMEDOUT;
        return -1;
      } else if(strncmp(CONNECT_SUCCESS, status_buf, 
                        strlen(CONNECT_SUCCESS)) == 0) {
        return 0;
      } else {
        fprintf(stderr, "amphibian_app: (bug) connect unpredictably: (%d) '%s'\n", 
                readbytes, status_buf);
        errno = EISCONN; /* a possible problem. */
        return -1;
      }
    } else {
      return connect(sockfd, serv_addr, addrlen);
    }
  } else {
    fprintf(stderr, "amphibian_app: null serv_addr passed to fishsockets_connect\n");
    abort();
  }
}

/* returns 0 on success, -1 on failure */
int fishsockets_bind(int sockfd, struct sockaddr *my_addr, socklen_t addrlen) {
  /* HORRIBLY incomplete -- this grabs a *specified* protocol socket. */
  /*   requires my_addr.sa_family to be set to AF_FISH if we're going to use it! */
  /*   does not support - zero values for my_addr.sfish_protocol (not allocated for you) */
  /*   does not support - non-zero values for my_addr.sfish_address (ignored)  */
  if( ((struct sockaddr_fish *)my_addr)-> sfish_family == AF_FISH ) {
    if( addrlen == sizeof(struct sockaddr_fish)) {
      /* TODO: make this handshake as well -- bind should
         perhaps fail if the port is already in use... */
      struct connection_request req;
      req.connect_not_bind = FALSE;
      memcpy(&req.dst, my_addr, addrlen);
      return write(sockfd, &req, sizeof(req));
    } else {
      return -1;
    }
  } else {
    return bind(sockfd, my_addr, addrlen);
  }
}

/* Only prepared to rcv one fd per message */
static int rcvfd(int passfd, int *data, int *datalen)
{
  char cbuf[CMSG_SPACE(sizeof(int))]; 
  struct msghdr mh = {0,0,0,0,0,0,0};
  struct cmsghdr *cm;
  int ret; 
  struct iovec iov; 
  int fd;

  assert(passfd != -1);
  assert(data != NULL);
  assert(datalen != NULL);

  if (data) { 
    mh.msg_iov = &iov; 
    mh.msg_iovlen = 1;
    iov.iov_base = data; 
    iov.iov_len = sizeof(int); 
  }

#if defined(HAVE_ST_MSG_CONTROL) || defined(__linux__)
  mh.msg_control = cbuf;
  mh.msg_controllen = sizeof cbuf; 
  cm = CMSG_FIRSTHDR(&mh); 
  cm->cmsg_len = CMSG_LEN(sizeof(int)); 
  cm->cmsg_level = SOL_SOCKET;
  cm->cmsg_type = SCM_RIGHTS; 
#else
  mh.msg_accrights = (caddr_t)&fd;
  mh.msg_accrightslen = sizeof(fd);
  fd = -1;
#endif

  *datalen = 0; 
  ret = recvmsg(passfd, &mh, 0); 
  if (ret < 0) 
    return ret; 
  
#if defined(HAVE_ST_MSG_CONTROL) || defined(__linux__)
  if (datalen) *datalen = ret; 
  fd = *(int *)CMSG_DATA(cm); 
#endif
  return fd;
}

/* returns the accepted socket */
int fishsockets_accept(int sockfd, struct sockaddr *remote_addr, socklen_t *addrlen) {

  if(sockfd == -1) {
    fprintf(stderr, "amphibian_app: %s called on a bad socket\n", __FUNCTION__);
    errno = EBADF;
    return -1;
  }
  
  if( ((struct sockaddr_fish *)remote_addr)-> sfish_family == AF_FISH ) {
    if( *addrlen == sizeof(struct sockaddr_fish)) {
      int jenny_jenny = 2525, len = 4;
      int accepted_socket = rcvfd(sockfd, &jenny_jenny, &len);
      if(accepted_socket < 0) {
        fprintf(stderr, "amphibian_app: (bug!) accepted invalid socket: %d: %s\n", 
                accepted_socket, strerror(errno));
        return -1;
      }
      if(jenny_jenny != 8675309) {
        fprintf(stderr, "amphibian_app: (potential bug) who can I turn to? %d=0x%x (fd=%d)\n", 
                jenny_jenny, jenny_jenny, accepted_socket);
        fprintf(stderr, "amphibian_app: it appears that data was passed on an accepting socket.\n");
        // return -1;
      }
      if( read(accepted_socket, remote_addr, *addrlen) < (int)sizeof(struct sockaddr_fish)) {
        fprintf(stderr, "amphibian_app: read of accepted socket remote address failed.\n");
        return -1;
      }
      return accepted_socket;
    } else {
      return -1;
    }
  } else {
    return accept(sockfd, remote_addr, addrlen);
  }
}
