/* solaris */
#define BSD_COMP
#define _GNU_SOURCE
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <strings.h> /* solaris */
#include <sys/ioctl.h> /* solaris */
#include <stdlib.h>
#include <netinet/in.h>
#include <netdb.h>
#include <net/if.h>
#include <time.h>
#include <ctype.h>
#include <assert.h>
#include <fcntl.h>
#include "fishguts.h"
#include "fish.h"

#define MAX_NEIGHBORS 12
#define MAXPACKET 1024
#define MINIMUM_TIME_BETWEEN_TWO_SENDS .01

#ifndef _BIG_ENDIAN 
#ifdef __sun__
#error "sanity check failed: suns aren't big endian (remove this if false)"
#endif
#endif

#define min(x,y) ({ \
 const typeof(x) _xi = (x);       \
 const typeof(y) _yi = (y);       \
 (void) (&_xi == &_yi);           \
 _xi < _yi ? _xi : _yi; })
#define max(x,y) ({ \
 const typeof(x) _xa = (x);       \
 const typeof(y) _ya = (y);       \
 (void) (&_xa == &_ya);           \
 _xa > _ya ? _xa : _ya; })

#define InRange(low, val, up) (((low) <= (val)) && ((val) <= (up)))
#define PRINT_AND_BAIL(x) do { fish_debugframe(FISH_DEBUG_NONE, "Error", \
		                               pkt, frame_length); \
                               errorabort(x); } while(0);
#define offsetof(type, member) ((size_t)(&((type *)0)->member))

/*@-booltype boolean@*/
typedef int boolean;
#ifndef FALSE
#define FALSE (0)
#define TRUE (!(FALSE))
#endif

static char *fishhead_host = NULL;
static int fishhead_port = -1;
static int self_address = -1;

static boolean init = FALSE;
/* the sockets are -1 so that join can test whether they're
   valid to protect against being called twice */
static int sockfishhead = -1;
static int sockfishbits = -1;
static struct sockaddr_in headaddr;
static struct sockaddr_in bitsaddr;
static time_t last_send_time;

/* The different network topologies that a Fishnet node can join.  */
#define TOPOLOGY_MANAGED 0
#define TOPOLOGY_ADHOC 1
static int network_topology = -1;
static struct sockaddr_in broadcast_address; // broadcast address for the interface

/*
 * Maintain a list of fishnet addr -> ip addr for fishguts to use to send
 * packets (instead of broadcasting all packets).
 */
struct mapping_entry *mapping_table = NULL;


/* +1 to allow null termination */
static struct neighbor neighbors[MAX_NEIGHBORS+1];

static receive_handler recvhandler;
static keyboard_handler keybhandler;

static void read_neighbor_list(int sd_fishhead);
static void frame_verification(const struct packet *frame, int len);

static int fish_constructmask(fd_set *prd);
static void fish_processbindings(fd_set *prd);

/* neighbor history routines; don't fail on a send
   frame if the neighbor was there before. */
#define MAX_NEIGHBOR_HIST 1024
static int neighbor_hist[MAX_NEIGHBOR_HIST];

boolean unsafe = TRUE;

#define MALLOC_Z(bytes) memset(malloc(bytes), 0, bytes)

static int find_in_neighbor_history(int naddr) {
  int i;
  for(i=0; i<MAX_NEIGHBOR_HIST
        && neighbor_hist[i]!= 0
        && neighbor_hist[i]!= naddr ;
      i++);
  if(neighbor_hist[i] == naddr)
    return i;
  else
    return -1;
}

static void insert_neighbor_history(int naddr) {
  int i;
  for(i=0; i<MAX_NEIGHBOR_HIST
        && neighbor_hist[i]!= 0
        && neighbor_hist[i]!= naddr ;
      i++);
  if(i < MAX_NEIGHBOR_HIST) {
    if(neighbor_hist[i] == 0)
      neighbor_hist[i] = naddr;
  } else {
    printf("Non-fatal error: ran out of neighbor history space.\n");
  }
}

static void
update_neighbor_history(void) {
  int i;
  for (i=0; i< MAX_NEIGHBORS && neighbors[i].address != 0 ; i++);
  insert_neighbor_history(neighbors[i].address);
}


/* used when the likely cause is system (possibly TA) error */
static void
errorexit(const char* msg, int syserr) {
  fprintf(stderr, "%s\n",msg);
  if (syserr != 0) perror("\n fish ");
  exit(EXIT_FAILURE);
}

/* used when the likely cause is programmer (read: student) error */
void
errorabort(const char* msg) {
  fprintf(stderr, "** %s **\n", msg);
  fprintf(stderr, "Generating core dump.  Use: \n"
          "  gdb [programname] core\n"
          "and use the commands: where, up, down, list, and print,\n"
          "if the error messages above (in **'s) haven't helped.\n");

  /* bail, generating core dump (if not limited, see getrlimit(2)) */
  abort();
}

static void
fish_init(void) {
  (void)time(&last_send_time);
  init = TRUE;
  srandom((unsigned int)time(0));

  return;
}

/* delivers a message */
static void deliver(int fromid, char *buf, int msglen) {

  char msg[30];
  sprintf(msg, "fish: received frame from %d", fromid);
  fish_debugframe(FISH_DEBUG_ALL, msg, buf, msglen);
  if (recvhandler != NULL)
    (*recvhandler)(fromid, buf, msglen);
}

/*
 * Search through the global mapping list to determine if the fishnet
 * addr is already contained. If the fishaddr is found, return 1; otherwise
 * return -1. Set the incoming pointers to the entry itself, and to the
 * entry preceding it.
 */
static int
find_fishaddr_mapping(int fishaddr, struct mapping_entry **entry)
{
  struct mapping_entry *p;

  assert(entry);
  *entry = NULL;

  /* XXX: why is this not (fishaddr <= 0) ? -ns */
  if( fishaddr < 0 )
    return -1;

  // Search through the list until you either find the fishaddr or the end
  for( p=mapping_table; p != NULL; p = p->next ) {

    // If we've found the address, set the entry and return
    if( p->fish_addr == fishaddr ) {
      *entry = p;
      return 1;
    }
  }
  return -1;
}


/*
 * Maintain a cache of ip addr -> fish addr pairs. This is then used when
 * sending packets to determine the ip address that they should be sent to.
 */
static void
update_mapping_cache(int fish_addr, struct sockaddr_in *from)
{
  struct mapping_entry *entry;

  assert(from->sin_addr.s_addr != broadcast_address.sin_addr.s_addr);

  if( find_fishaddr_mapping(fish_addr, &entry) > 0 ) {
    // If the fishaddr is already in the mapping_table, then just check to
    // make sure that the IP addr hasn't changed. If it has, change the mapping
    // and print out a warning.

    if( entry->node_ip_address.s_addr != from->sin_addr.s_addr ) {
      fish_debug(FISH_DEBUG_NONE,
                 "fish: fishnet node %d has moved from %s to %s\n",
                 fish_addr,
                 inet_ntoa(entry->node_ip_address),
                 inet_ntoa(from->sin_addr));
      entry->node_ip_address = from->sin_addr;
    }
  } else {
    // The fishaddr was not found, so we need to add it to our list.

    struct mapping_entry *node = malloc(sizeof(struct mapping_entry));
    if( !node ) {
      fish_debug(FISH_DEBUG_INTERNAL,
                 "fish: error, cannot allocate memory for mapping node\n");
      return;
    }
    node->fish_addr = fish_addr;
    node->node_ip_address = from->sin_addr;

    // Add to the global mapping table
    node->next = mapping_table;
    mapping_table = node;
  }
}


/*
 * Get the appropriate ip addr for the provided fish addr. The sockaddr_in
 * struct is filled in if we have the mapping. If the destination is not
 * known, then the sockaddr is filled in with a broadcast address.
 * Return 1 on success; -1 on failure.
 */
static int
get_ip_mapping(int fish_addr, struct sockaddr_in *to)
{
  struct mapping_entry *entry = NULL;

  if( !to || (fish_addr < 0) ) {
    return -1;
  }

  memset((void *)to, 0, sizeof(struct sockaddr_in));
  to->sin_family      = AF_INET;
  to->sin_port        = htons(FISHBITS_BCST_PORT);

  // If we find an IP address, then we'll only send the packet to that IP
  // address. Otherwise, we'll broadcast the packet to all within listening
  // distance.
  if( find_fishaddr_mapping(fish_addr, &entry) > 0 ) {
    to->sin_addr = entry->node_ip_address;

  } else {
    to->sin_addr = broadcast_address.sin_addr;
  }

  return 1;
}


/*
 * Process the fishguts packet. First, check that the major version is
 * compatible with our version. Next, update the arp cache based upon
 * the intermediate nodes IP addresses. Finally, determine if this
 * packet should be sent up to the user, setting the fromid to the
 * intermediate fish_src address. Return 1 if the packet should be deliver,
 * -1 if it should be dropped.
 */
static int
process_fishguts_packet(struct fishguts_packet *p_pkt,
                        struct sockaddr_in *from,
                        int msglen,
                        int *fromid)
{
  if( !p_pkt || !from | !fromid )
    return -1;

  // set it to something bogus just in case we return with an error
  *fromid = -1;

  if( msglen < (int)(sizeof(struct fishguts_header)) )
    return -1;

  // is the packet from a fishguts of the same major version?
  if( p_pkt->fg_hdr.ver.v_major != FISHGUTS_VERSION_MAJOR ) {
    fish_debug(FISH_DEBUG_NONE,
               "fish: error, packet received with incorrect version\n");
    return -1;
  }

  // Drop broadcast packets that were sent by this fishnet node
  // Note that this is consistent with what the managed version of
  // send to ALL_NEIGHBORS does.
  if( (p_pkt->fg_hdr.fish_dst == ALL_NEIGHBORS) &&
      (p_pkt->fg_hdr.fish_src == self_address) ) {
    return -1;
  }

  // update the internal cache of ip addr->fish addr if in adhoc mode
  if( network_topology == TOPOLOGY_ADHOC ) {
    update_mapping_cache(p_pkt->fg_hdr.fish_src, from);
  }

  // determine if this packet is destined for the user
  if( (p_pkt->fg_hdr.fish_dst == ALL_NEIGHBORS) ||
      (p_pkt->fg_hdr.fish_dst == self_address) ) {
    *fromid = p_pkt->fg_hdr.fish_src;
    return 1;
  }

  return -1;
}



/* handles input from other fishnet nodes */
static void
receive_handler_wrapper(int sd_fishbits __attribute__((unused))) {
  struct fishguts_packet *p_pkt = 
    (struct fishguts_packet *)MALLOC_Z(MAX_FISHGUTS_PKT_SIZE+1);
  int msglen;
  struct sockaddr_in from;
  socklen_t fromlen = sizeof(struct sockaddr_in);
  int fromid;

  msglen = recvfrom(sockfishbits, p_pkt, MAX_FISHGUTS_PKT_SIZE,
                    0, (struct sockaddr*)&from, &fromlen);

  // Deliver the packet if it is destined for us (either to us or broadcast)
  if( process_fishguts_packet(p_pkt, &from, msglen, &fromid) > 0 ) {
    if( fromid < 0 ) {
      fish_debug(FISH_DEBUG_NONE,
                 "fish: error, received packet with invalid fishguts src\n");
    } else {
      int difflen = offsetof(struct fishguts_packet, pkt);
      deliver(fromid, (char *)&p_pkt->pkt, msglen - difflen);
    }
  }
  free(p_pkt);
}


void
fish_recvhook(receive_handler fn) {
  if (!init) fish_init();
  recvhandler = fn;
}

void
keyboard_handler_wrapper(int stdin_fd) {
  char buf[1024];

  fgets(buf,sizeof(buf),stdin);
  if(feof(stdin)) {
    fish_debug(FISH_DEBUG_NONE, "fish: stdin closed, type 'exit' instead of ^D, or hit ^C now to quit.\n");
    fish_remove_readhook(stdin_fd);
    /* exit(EXIT_SUCCESS); */
    return;
  }
  if(strncmp(buf, "exit", 4)==0) {
    fish_debug(FISH_DEBUG_NONE, "fish: exiting\n");
    exit(EXIT_SUCCESS);
  }

  fish_debug(FISH_DEBUG_ALL, "fish: received keyboard input\n");
  if (keybhandler != NULL)
    (*keybhandler)(buf);
}

void
fish_keybhook(keyboard_handler fn) {
  if (!init) fish_init();
  fish_readhook(fileno(stdin), keyboard_handler_wrapper);
  keybhandler = fn;
}

/* sends a frame to a udp port, ip address */

static int
udpsend(struct sockaddr_in to, int naddr,  void* frame, int len) {
  int ok;

  ok = sendto(sockfishbits, frame, len, 0, (struct sockaddr*)&to, sizeof(to));
  if (ok < 0) {
    if (errno == ECONNREFUSED) {
      fish_debug(FISH_DEBUG_TOPOLOGY,
                 "fish: warning - neighbor %d seems to have failed\n",
                 naddr);
    } else {
      fish_debug(FISH_DEBUG_NONE,
                 "fish: while sending to %d at %s:%d\n",
                 naddr, inet_ntoa(to.sin_addr), htons(to.sin_port));
      fish_debug(FISH_DEBUG_NONE,
                 "fish: experienced error %d: %s\n",
                 errno, strerror(errno));
      errorexit("fish: error, can't send\n", 1);
    }
  }

  // limit sending rate
  if(!unsafe) usleep(10);

  return ok;
}

/*
 * Actually send a frame to a neighbor; called by fish_send
 * The result is <0 on failure.
 */
static int
realsend(int n, int naddr, void* frame, int len) {
  struct sockaddr_in to;

  if ((unsigned int)((random() >> 16 | random()) % 1000)
      < (unsigned int)neighbors[n].lossrate_x_1000) {
    fish_debug(FISH_DEBUG_ALL,
               "fish: Dropped frame from %d to %d\n",
               self_address, naddr );
    /* should be a silent failure. */
    return 0;
  }

  memset((void *)&to, 0, sizeof(to));
  to.sin_family      = AF_INET;
  to.sin_addr        = neighbors[n].node_ip_address;
  to.sin_port        = neighbors[n].node_udp_port;
  return udpsend(to,naddr,frame,len);
}


/*
 * Send a packet in unmanaged mode by searching through the mapping
 * table for the ip address associated with the fish addr. If no
 * ip address is found, this packet will be sent via broadcast.
 * The result is <0 on failure.
 */
static int
unmanagedsend(int fish_addr, int naddr, void* frame, int len)
{
  struct sockaddr_in to;

  // search through the ip addr -> fish addr mapping list
  get_ip_mapping(fish_addr, &to);

  return udpsend(to,naddr,frame,len);
}


/*
 * Add a fishguts header to the user-defined packet frame. Set the
 * major and minor number.
 */
static int
add_fishguts_header(void *frame, int len, struct fishguts_packet *fg_pkt)
{
  int fg_len;

  if( !frame || !fg_pkt )
    return -1;

  memset((void *)&fg_pkt->pkt, 0, sizeof(struct packet));
  memcpy((void *)&fg_pkt->pkt, (const void *)frame, len);
  fg_pkt->fg_hdr.ver.v_major = FISHGUTS_VERSION_MAJOR;
  fg_pkt->fg_hdr.ver.v_minor = FISHGUTS_VERSION_MINOR;
  fg_len = len + offsetof(struct fishguts_packet, pkt);

  return fg_len;
}


/*
 * Send a packet over a managed network. This typically means that there
 * is a fishhead dictating the network topology, rather than just the
 * physical environment.
 */
static int
fish_send_managed(int naddr, struct fishguts_packet *fg_pkt, int fg_len)
{
  int n;
  int i;
  struct sockaddr_in to;
  int status = -1;

  switch (naddr) {

  case ANY_NEIGHBOR:
    // anycast
    for (i=0; (i < MAX_NEIGHBORS) && (neighbors[i].address != 0); i++);
    if (i==0) {
      fish_debug(FISH_DEBUG_NONE,
                 "fish: no neighbors to send to in ANYCAST\n");
      return -1;
    } else {
      int neighbor = random() % i;
      fg_pkt->fg_hdr.fish_dst = neighbor;
      status = realsend(neighbor, naddr, (void *)fg_pkt, fg_len);
    }
    break;
  case ALL_NEIGHBORS:
    // broadcast
    for (i=0; (i < MAX_NEIGHBORS) && (neighbors[i].address != 0); i++) {
      fg_pkt->fg_hdr.fish_dst = ALL_NEIGHBORS;
      status = realsend(i, naddr, (void *)fg_pkt, fg_len);
    }
    break;

  default:

    // local loopback delivery
    if (naddr == self_address) {
      memset((void *)&to, 0, sizeof(to));
      to.sin_family      = AF_INET;
      to.sin_addr.s_addr = inet_addr("127.0.0.1");
      to.sin_port        = bitsaddr.sin_port ;
      fg_pkt->fg_hdr.fish_dst = self_address;
      status = udpsend(to,self_address,(void *)fg_pkt,fg_len);
    } else {
      // neighborcast
      n = -1;
      for (i = 0; i < MAX_NEIGHBORS; i++) {
        if (neighbors[i].address == naddr) n = i;
      }
      if (n < 0) {
        /* only abort if this neighbor never existed. */
        if(find_in_neighbor_history(naddr) == -1) {
          fish_debug(FISH_DEBUG_NONE,
                     "fish: %d not a valid neighbor\n",
                     naddr);
          return -1;
        } else {
          errorabort("fish: not a valid neighbor\n");
        }
      }
      fg_pkt->fg_hdr.fish_dst = naddr;
      status = realsend(n, naddr, (void *)fg_pkt, fg_len);
    }
  }
  return status;
}



/*
 * Send a packet over an unmanaged network. This typically means that there
 * is not a fishhead dictating the network topology, rather the topology is
 * just the physical environment.
 */
static int
fish_send_unmanaged(int naddr, struct fishguts_packet *fg_pkt, int fg_len)
{
  struct sockaddr_in to;
  int status = -1;

  switch (naddr) {

  case ANY_NEIGHBOR:
    // Don't think it is support in unmanaged mode. You don't have a
    // notion of neighbors, so you can't just randomly pick a fish addr
    // and send a packet.
    fish_debug(FISH_DEBUG_NONE,
               "fish: error, anycast not supported in unmanaged mode\n");
    return -1;
    break;

  case ALL_NEIGHBORS:
    // broadcast
    fg_pkt->fg_hdr.fish_dst = naddr;
    status = unmanagedsend(0, naddr, fg_pkt, fg_len);
    break;

  default:

    // local loopback delivery
    if (naddr == self_address) {
      memset((void *)&to, 0, sizeof(to));
      to.sin_family      = AF_INET;
      to.sin_addr.s_addr = inet_addr("127.0.0.1");
      to.sin_port        = bitsaddr.sin_port ;
      fg_pkt->fg_hdr.fish_dst = self_address;
      status = udpsend(to,self_address,(void *)fg_pkt,fg_len);
    } else {

      fg_pkt->fg_hdr.fish_dst = naddr;
      status = unmanagedsend(naddr, naddr, (void *)fg_pkt, fg_len);
    }
  }
  return status;
}



int
fish_send(int naddr, void* frame, int len) {
  struct fishguts_packet fg_pkt;
  int fg_len;
  int status = -1;

  if (!init) fish_init();

  if(sockfishbits < 0) {
    errorabort("fish: must join the network before sending\n");
  }

  if (len < 1) {
    errorabort("fish: zero-length frame\n");
  }

	
  if (len > MAXPACKET) {
    errorabort("fish: frame too long\n");
  }

  frame_verification(frame,len);

  // Create a frame with the fishguts header. Set the major and minor version.
  fg_len = add_fishguts_header(frame, len, &fg_pkt);
  if( fg_len < 0 ) {
    fish_debug(FISH_DEBUG_NONE,
               "fish: error, couldn't add fishguts header\n");
    return -1;
  }

  // Set the fishguts source address to our address
  fg_pkt.fg_hdr.fish_src = self_address;

  if( network_topology == TOPOLOGY_ADHOC ) {
    status = fish_send_unmanaged(naddr, &fg_pkt, fg_len);
  } else {
    status = fish_send_managed(naddr, &fg_pkt, fg_len);
  }

  if (status >= 0) {
    fish_debugframe(FISH_DEBUG_ALL, "fish: Sent frame", frame, len);
    return 0;
  }

  return -1;
}


void parse_location(const char* location, char** host, int* port) {

  char* separator = index(location, ':');
  int hostlength = separator - location;
  if (separator == NULL) {
    errorabort("fish: invalid port number\n");
  }

  *host = malloc(hostlength+1);
  strncpy(*host, location, hostlength);
  (*host)[hostlength] = '\0';

  *port = atoi(separator+1);
}


/*
 * Prepare for UDP broadcast by getting the interface broadcast address and
 * setting the broadcast capabilities. This function terminates the program
 * if it fails.
 *
 * Note that we assume that the first non-loopback device is the device that
 * is being used for wireless. This should be fixed. FIXME MAR
 */
static void
init_broadcast(int sfd)
{
  // Set the socket options so that it can send and receive broadcast packets
  {
    int optval = 1;
    if( setsockopt(sfd, SOL_SOCKET, SO_BROADCAST,
                   (void *)&optval, sizeof(optval)) < 0 ) {
      errorexit("fish: error, cannot broadcast in adhoc network\n", 1);
    }
  }

  // Get the broadcast address for this interface, just in case we need it
  {
    struct ifreq ifr;
    struct ifconf ifc;
    struct ifreq ifaces[12];
    int foundIf = 0;
    int num_interfaces, i;

    // FIXME: For the moment, we're assuming that there is only a single
    // interface being used in the wireless device.
    ifc.ifc_len = sizeof(ifaces);
    ifc.ifc_buf = (char *)ifaces;

    if( ioctl(sfd, SIOCGIFCONF, &ifc) < 0 ) {
      errorexit("fish: error, couldn't get network interface\n", 1);
    }

    num_interfaces = ifc.ifc_len / sizeof(struct ifreq);

    for( i = 0; i< num_interfaces; i++ ) {
      if( (ifaces[i].ifr_addr.sa_family == AF_INET) &&
          ( strncmp(ifaces[i].ifr_name, "lo", strlen("lo")) ) ) {
        foundIf = 1;
        strncpy(ifr.ifr_name, ifaces[i].ifr_name, sizeof(ifr.ifr_name));
        break;
      }
    }

    // If we didn't find an appropriate interface, we need to croak
    if( !foundIf ) {
      errorexit("fish: error, no interface to broadcast on\n", 1);
    }

    if( ioctl(sfd, SIOCGIFBRDADDR, &ifr) < 0 ) {
      errorexit("fish: error, couldn't get broadcast address\n", 1);
    } else {
      if( ifr.ifr_broadaddr.sa_family == AF_INET ) {
        struct sockaddr_in *sin_p = (struct sockaddr_in *)&ifr.ifr_broadaddr;
        broadcast_address.sin_addr = sin_p->sin_addr;
      } else {
        errorexit("fish: error, wrong family for broadcast address\n", 1);
      }
    }
  }
}



/*
 * Based upon the passed in string, determine if we are connecting to an adhoc
 * or a managed network. Set the global variable network_topology.
 */
static void
set_network_topology(const char *str)
{
  if( str == NULL ) {
    network_topology = TOPOLOGY_ADHOC;
  } else if( !strcmp(str, "") ) {
    network_topology = TOPOLOGY_ADHOC;
  } else {
    network_topology = TOPOLOGY_MANAGED;
  }
}

/* for code that tests subsystems that do not require a
   network, but do require fish_main() */
void fish_join_fake_network(void) { sockfishbits = 0; }

/* helper for join network, deal with sockfishbits */
static void initialize_fishbits_socket(void) {
  int ok;
  socklen_t address_length_buf;
  // Open a UDP Port
  sockfishbits = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);
  if (sockfishbits < 0)
    errorexit("fish: error,  can't get fishbits socket\n", 1);
  
  bzero((char *) &bitsaddr, sizeof(bitsaddr));
  bitsaddr.sin_family      = AF_INET;
  bitsaddr.sin_addr.s_addr = htonl(INADDR_ANY);
  if( network_topology != TOPOLOGY_ADHOC ) {
    bitsaddr.sin_port        = htons(0);
  } else {
    bitsaddr.sin_port        = htons(FISHBITS_BCST_PORT);
  }
  
  ok = bind(sockfishbits, (struct sockaddr*)&bitsaddr, sizeof(bitsaddr));
  if (ok < 0)
    errorexit("fish: error, can't bind fishbits socket\n", 1);
  
  // Bind the socket to a handler off the main select() loop
  fish_readhook(sockfishbits, receive_handler_wrapper);
  
  // Get the port number allocated by the system, so that
  // we can tell fishhead how others can contact us.
  address_length_buf = sizeof(bitsaddr);
  ok = getsockname(sockfishbits,
                   (struct sockaddr *)&bitsaddr,
                   &address_length_buf);
  if (ok < 0)
    errorexit("fish: error, can't find port number of fishbits socket\n", 1);
  
}

static void initialize_fishhead_socket(int fish_addr) {
  int zok;
  struct hostent *hp;
  struct in_addr *firstAddress;
  struct fish_message *pfm;
  struct joinrequest *jreq;

  /// Code handling the network manager
  sockfishhead = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
  if (sockfishhead < 0)
    errorexit("fish: error, can't get fishhead socket\n", 0);

  // Convert hostname to  ipaddress
  if ( (hp = gethostbyname(fishhead_host)) == NULL) {
    fprintf(stderr,"gethostbyname() failed trying to resolve '%s'\n",
            fishhead_host);
    errorexit(" Failure in mapping host name to IP",1);
  }
  firstAddress = (struct in_addr *) hp->h_addr_list[0];
  if (firstAddress == NULL)
    errorexit(" Failure in mapping host name to IP",0);
  bzero((char *) &headaddr, sizeof(headaddr));
  memcpy(&headaddr.sin_addr, firstAddress, sizeof(struct in_addr));
  headaddr.sin_family      = AF_INET;
  headaddr.sin_port        = htons(fishhead_port);

  // Connect to the network manager
  zok = connect(sockfishhead,
                (struct sockaddr*) &headaddr,
                sizeof(headaddr));
  if (zok < 0)
    errorexit("fish: error, can't get fishhead connection\n", 1);

  // Send a join request to the network manager
  pfm = (struct fish_message *)malloc(1024);
  pfm->type   = FISH_MSG_TYPE_JOINREQUEST;
  pfm->length = MIN_MESSAGE_SIZE + (sizeof(struct joinrequest));
  jreq = &pfm->body.joinreq;
  jreq->node_address = fish_addr;
  jreq->node_udp_port = bitsaddr.sin_port;
  strncpy(jreq->username, getenv("USER"), 8);

  if(write(sockfishhead, pfm, (size_t)pfm->length) < 0) {
    errorexit(" Failure in writing join request to server",1);
  }
  free(pfm);
}

void fish_joinnetwork(const char* fishhead_location, int addr) {

  if (sockfishhead > 0 || sockfishbits > 0)
    errorabort("fish: error, already joined\n");

  if ( ! InRange(1, addr, MAX_NODE_ADDRESS)) {
    errorexit("fish: not a valid fishnet address", 0);
  }

  if (!init) fish_init();

  // Determine and set the network topology
  set_network_topology(fishhead_location);

  // parse the fishhead location
  if( network_topology != TOPOLOGY_ADHOC ) {
    parse_location(fishhead_location, &fishhead_host, &fishhead_port);
  }
  self_address = addr;

  initialize_fishbits_socket();

  // If we're in adhoc mode, we have no need for a network manager
  if( network_topology == TOPOLOGY_ADHOC ) {
    init_broadcast(sockfishbits);
    
    fish_debug(FISH_DEBUG_TOPOLOGY,
               "fish: joined adhoc network as %i\n",
               self_address);	
  } else {
    initialize_fishhead_socket(addr);
    
    /* Now wait to hear the return message from the server
     * telling you your neighbor list... */
    (void)read_neighbor_list(sockfishhead);
    printneighbors(neighbors);
    
    /* setup input handler to get neighbor list updates from fishhead */
    fish_readhook(sockfishhead, read_neighbor_list);
    
    fish_debug(FISH_DEBUG_TOPOLOGY,
               "fish: joined network %s:%i as %i\n",
               fishhead_host, fishhead_port, self_address);
  }
}

const char* fish_getheadhost(void) {
  return fishhead_host;	
}

int fish_getheadport(void) {
  return fishhead_port;
}

int fish_getaddress(void) {
  return self_address;
}

void fish_main(void) {

  if (!init) fish_init();

  if(sockfishbits<0) {
    errorabort("fish: must join the network before calling fish_main\n");
  }

  fish_debug(7, "fish: in main\n");

  while (1) {
    fd_set rd;
    int max = 0;
    struct timeval wait;
    int ok;

    max = fish_constructmask(&rd);

    /* patch for apparent problem with stdin.  I don't
       understand why this is necessary, but fixing it this
       way is better than leaving it broken. */
    if(FD_ISSET(fileno(stdin), &rd) && fcntl(fileno(stdin), F_GETFL) < 0) {
      fish_remove_readhook(fileno(stdin));
      max = fish_constructmask(&rd);
    }

    ok = select(max+1, &rd, 0, 0,
                fish_time_to_next_event(&wait) ? &wait : 0);
    if (ok < 0) {
      if(errno != EINTR) {
        if(errno == EBADF) {
          int i;
          printf("Bad file descriptor error: be careful to call fish_remove_readhook() with close()\n");
          for(i=0; i<max; i++) {
            if(FD_ISSET(i, &rd)  && fcntl(i, F_GETFL) < 0) {
              printf("  file descriptor #%d is suspicious: %s\n",
                     i, strerror(errno));
            }
          }

        } else {
          perror("select");
        }
        errorabort("fish: error calling select()");
      } else {
        continue;
      }
    }

    /* do scheduled things */
    fish_executepending();

    /* handle anything that uses sockets, including
       fishhead, fishbits, and keyboard */
    fish_processbindings(&rd);
  }
}

/* the same as GNU strnlen, portable because we rewrite it. */
static unsigned int 
our_strnlen(const char *string, unsigned int max_len) {
  unsigned int ret;
  for(ret = 0; ret < max_len && string[ret] != '\0'; ret++);
  return(ret);
}

/* helper function for frame_verification (below) that
   verifies that the length of a text containing frame
   (either assignment 1 style or ECHO/ECHO_RESPONSE is no
   larger than necessary */
static boolean
checkStringLength(const char *string,
                  unsigned int string_length) {
  if(string_length != 0 && our_strnlen(string, string_length) < string_length-1) {
    fprintf(stderr, "The frame containing the following text:\n"
            "%*s\n"
            "was sent with a length of %u bytes, "
            "but only requires %d bytes\n",
            string_length, string,
            string_length, our_strnlen(string,string_length));
    return(FALSE);
  } else
    return(TRUE);
}


/*
 * Use the cksum function from Petersen & Davie Section 2.4.2. The checksum
 * can either be zero, or it must use this format. The checksum is calculated
 * with both the ttl and the checksum set to 0.
 *
 * Return 1 if the packet checksum is correct or 0; return -1 otherwise.
 */
static unsigned short
cksum(const unsigned short *buf, int count)
{
  register unsigned long sum  = 0;

  while (count--)
    {
      sum += *buf++;
      if (sum & 0xFFFF0000)
        {
          /* carry occurred, so wrap around */
          sum &= 0xFFFF;
          sum++;
        }
    }
  return ~(sum & 0xFFFF);
}

static int
verify_checksum(const struct packet *pkt, int frame_length, int methodical)
{
  // Zero is an acceptable checksum value
  if( pkt->hdr.cksum == 0 )
    return 1;
    
  if(!methodical) {  
    /* fast, in place check, but does not say what the
       checksum value should have been */

    /* calculate the checksum over the whole packet as it stands */
    int should_be_xffff = cksum((const unsigned short *)pkt, (frame_length)/2);

    /* the last byte needs to be "added" to the ones complement negative sum */
    if(frame_length % 2) {
      unsigned char last_byte = ((const unsigned char *)pkt)[frame_length - 1];
      
#ifdef _BIG_ENDIAN 
      should_be_xffff += ~((unsigned short)(last_byte) << 8) + 1;
#else
      should_be_xffff += ~(unsigned short)last_byte + 1;
#endif
      /* and carry dealt with */
      should_be_xffff += should_be_xffff >> 16;
      should_be_xffff &= 0xffff;
    }

    /* the ttl needs to be "subtracted" from the ones complement, because it
       didn't go into the checksum to begin with */
    should_be_xffff += (pkt->hdr.ttl) ? pkt->hdr.ttl : 0xffff;

    /* and carry dealt with */
    should_be_xffff += should_be_xffff >> 16;
    should_be_xffff &= 0xffff;

    if( should_be_xffff == 0xffff ) {
      return 1;
    } else {
      printf("Should be 0xffff is actually %d(0x%x) packet checksum: %d(0x%x) last byte=(0x%x) len=%d\n",
             should_be_xffff, should_be_xffff, 
             pkt->hdr.cksum, pkt->hdr.cksum,
             ((const unsigned char *)pkt)[frame_length-1], frame_length);
      /* use the methodical test to show what it should have been */
      return verify_checksum(pkt, frame_length, TRUE);
    }
  } else {
    struct packet *copy_pkt = NULL;
    unsigned short calculated_sum = 0;
    int byte_len;
    
    // Make sure that we calculate on a 16 bit boundary
    byte_len = frame_length;
    if( (byte_len % 2) != 0 ) {
      byte_len = byte_len + 1;
    }
    
    // Just to be absolutely certain there isn't anything funky about the arch
    assert( ( (byte_len * 8 * sizeof(char)) % 16 ) == 0 );
    
    // Allocate a buffer large enough that is 16bit bound
    copy_pkt = malloc(byte_len * sizeof(char));
    if( copy_pkt == NULL ) {
      return -1;
    }
    
    // Clear the buffer & copy the packet and set the ttl and checksum to 0
    memset((void *)copy_pkt, 0, byte_len * sizeof(char));
    memcpy((void *)copy_pkt, (const void *)pkt, frame_length);
    copy_pkt->hdr.ttl = 0;
    copy_pkt->hdr.cksum = 0;
    
    // Calculate the checksum on this packet
    calculated_sum = cksum((unsigned short *)copy_pkt, (byte_len / 2));
    
    free(copy_pkt);
    
    // Nope. Try again please.
    if( (int)calculated_sum != pkt->hdr.cksum ) {
      printf("Calculated checksum: %d(%x) packet checksum: %d(%x)\n",
             calculated_sum, calculated_sum, 
             pkt->hdr.cksum, pkt->hdr.cksum);
      return -1;
    }
    
    return 1;
  }
}


/* abort()s on failure */
static void
frame_verification(const struct packet *frame, int frame_length) {
  int i;
  int textframe;
  int toCheckLen;
  const struct packet *pkt = (const struct packet *)frame;

  if(frame == NULL)
    errorabort("Null frame");

	// printf("frame_length = %i, MTU = %i\n", frame_length, MTU);
  if (frame_length > MTU)
    errorabort("Size of frame > MTU");
  if (frame_length < 0)
    errorabort("Frame has negative size");

  // Check if the frame is text only...
  /* this is old code and probably should be removed */
  toCheckLen = min(16, frame_length);
  for(i=0, textframe = 1; i< toCheckLen; i++) {
    int c = ((const char *)frame)[i];
    textframe &= (isprint(c) || isspace(c)) ;
  }

  if (textframe) {
    /* Verify that the frame length (the function parameter)
       is at most as long as the string plus a safe
       terminating zero.  This is not an egregious error,
       but verifying it here is probably the only way to
       make sure students pass a length parameter that is
       not too long. */
    if(!checkStringLength((const char *)frame, frame_length)) {
      errorabort("Text frame was too long");
    }
    return;
  }

  // Packet is not just text
  if ( ! InRange(1, pkt->hdr.src, MAX_NODE_ADDRESS)) {
    PRINT_AND_BAIL("In send: Invalid Source");
  }

  if ( ! InRange(-1, pkt->hdr.dst, MAX_NODE_ADDRESS)) {
    PRINT_AND_BAIL("In send: Invalid Destination");
  }

  // It has been determined that ALL_NEIGHBORS actually means flood the
  // network by djw. Therefore, nodes may forward "broadcast" packets without
  // changing the original fishnet source.
  //  if ( (pkt->hdr.dst == ALL_NEIGHBORS) && (pkt->hdr.src != self_address)) {
  //    PRINT_AND_BAIL("In send: src != self for packet sent to ALL_NEIGHBORS");
  //  }

  if ( ! InRange(1, pkt->hdr.ttl, MAX_TTL)) {
    PRINT_AND_BAIL("In send: Invalid TTL");
  }

  if ( ! InRange(1, pkt->hdr.protocol, FISH_PROTOCOL_MAX)) {
    PRINT_AND_BAIL("In send:Invalid Protcol");
  }

  // Check that the checksum is either set to 0 or that the checksum is correct
  if( verify_checksum(pkt, frame_length, FALSE) < 0 ) {
    PRINT_AND_BAIL("In send: Invalid checksum");
  }

  if (pkt->hdr.protocol == FISH_PROTOCOL_ROUTING_DV){
    if (pkt->hdr.ttl != 1)  {
      PRINT_AND_BAIL("In send: TTL must be 1 one for a Routing Packet");
    }
    if (pkt->hdr.src != self_address) {
      PRINT_AND_BAIL("In send: Source for a Routing Packet must be same as self");
    }
    
		
    // Check that the number of advertisements in the packet is consistent w/ length
    {
      struct routing_packet *rpkt = (struct routing_packet *)pkt;
      int adv_len, len_to_adv, diff_len;
      
      // Amount of bytes that should be taken up by routing advertisements
      adv_len = rpkt->route_hdr.num_routes * sizeof(struct route_advertisement);
      
      // Routing packet length up to the list of advertisements
      len_to_adv = offsetof(struct routing_packet, adv);

      // Calculate the difference between the routing packet frame length
      // and the size that it should be
      diff_len = frame_length - (len_to_adv + adv_len);
      
      if( diff_len < 0 ) {
        PRINT_AND_BAIL("In send: Routing Packet frame too small for enclosed "
                       "number of advertisements");
      } else if( diff_len > 0 ) {
        PRINT_AND_BAIL("In send: Routing Packet frame too large for enclosed "
                       "number of advertisements");
      }
    }
  }
	
  if (pkt->hdr.protocol == FISH_PROTOCOL_TRANSPORT){
    const struct transport_packet *tpkt =
      (const struct transport_packet *)frame;

    if ( ! InRange(0,tpkt->transport_hdr.src_port,MAX_PORT_NUMBER)) {
      PRINT_AND_BAIL("In send:Transport protocol packet has invalid src port");
    }
    if ( ! InRange(0,tpkt->transport_hdr.dst_port,MAX_PORT_NUMBER)) {
      PRINT_AND_BAIL("In send:Transport protocol packet has invalid dst port");
    }

    if(!((tpkt->transport_hdr.fin==0) || (tpkt->transport_hdr.fin ==1))) {
      PRINT_AND_BAIL("In send:Transport protocol packet has invalid fin (must be zero or one)");
    }

    if(!((tpkt->transport_hdr.syn==0) || (tpkt->transport_hdr.syn ==1))) {
      PRINT_AND_BAIL("In send:Transport protocol packet has invalid syn (must be zero or one)");
    }
  }

  if (pkt->hdr.protocol == FISH_PROTOCOL_ECHO) {
    struct echo_packet *epkt = (struct echo_packet *)pkt;
    if(!checkStringLength(epkt->data,
                          frame_length-
                          offsetof(struct echo_packet, data))) {
      PRINT_AND_BAIL("In send: ECHO/ECHO_RESPONSE frame was too long");
    }

    if( (epkt->echo_hdr.code != ECHO_PROTOCOL_REQUEST) &&
        (epkt->echo_hdr.code != ECHO_PROTOCOL_RESPONSE) ) {
      PRINT_AND_BAIL("In send: ECHO code not REQUEST or RESPONSE");
    }
  }
}


static void
read_neighbor_list(int sd_fishhead) {
#define MSG_BUF_SZ 1024
  struct fish_message *pfm = MALLOC_Z(MSG_BUF_SZ); 
  int readlen;

  readlen = read(sd_fishhead, pfm, MSG_BUF_SZ);
  if (readlen <= 0) {
    errorexit("Fishead died\n", (readlen < 0) ? 1 : 0);
  }

  fish_debug(FISH_DEBUG_TOPOLOGY, "Neighbor update received.\n");

  if(pfm->length > readlen) {
    errorexit("read() failed, message not completely read", 0);
  }

  if (pfm->type != FISH_MSG_TYPE_NEIGHBORLIST){
    printmessage(pfm);
    errorabort("Above Message received from fish-head \n");
  }
  else {
    int i;
    int num_neigh = (pfm->length - MIN_MESSAGE_SIZE) / sizeof(struct neighbor);
    // printmessage(pfm);
	
    for(i=0; i< num_neigh; i++) {
      neighbors[i].address = pfm->body.neighbors[i].address ;
      neighbors[i].node_ip_address =  pfm->body.neighbors[i].node_ip_address ;
      neighbors[i].node_udp_port = pfm->body.neighbors[i].node_udp_port ;
      neighbors[i].lossrate_x_1000 = pfm->body.neighbors[i].lossrate_x_1000;
      assert(neighbors[i].lossrate_x_1000<=1000);
    }
    for(i=num_neigh; i< MAX_NEIGHBORS; i++) {
      neighbors[i].address = 0;
    }
    update_neighbor_history();
  }
}

/* select bindings describe the handler to call
   when a socket becomes ready for writing */
static struct select_binding {
  struct select_binding *next;
  int sd;
  void (*handler)(int);
} *binding_list;

/* add a select binding */
void fish_readhook(int sd, void (*rdfn)(int)) {
  struct select_binding *psb =
    (struct select_binding *)malloc(sizeof(struct select_binding));
  if(sd < 0) { errorabort("bad socket passed to fish_readhook\n"); }
  if(psb == NULL) { errorabort("unable to allocate memory\n"); }
  psb->sd = sd;
  psb->handler = rdfn;
  psb->next = binding_list;
  binding_list = psb;
}

/* remove a select binding */
void fish_remove_readhook(int sd) {
  struct select_binding *psb, *qsb;
  /* printf("removing hook for %d\n", sd); */
  for(psb = binding_list, qsb = NULL;
      psb != NULL && psb->sd != sd;
      qsb=psb, psb=psb->next);

  if(psb != NULL) { /* found it */
    if(qsb != NULL) { /* in the middle */
      qsb->next = psb->next;
    } else { /* at the beginning */
      binding_list = psb->next;
    }
    free(psb);
  }
}

/* construct a select read mask from the select bindings */
static int fish_constructmask(fd_set *prd) {
  struct select_binding *psb;
  int max = 0;
  FD_ZERO(prd);
  for(psb = binding_list;
      psb != NULL;
      psb = psb->next) {
    FD_SET(psb->sd, prd);
    max = max(psb->sd, max);
  }
  return(max);
}

/* walk the bindings list and invoke the callbacks */
static void
fish_processbindings(fd_set *prd) {
  struct select_binding *psb, *next;
  for(psb = binding_list; psb != NULL; psb = next) {
    next = psb->next; /* grab this now, in case we remove
                         ourselves from the list */
    if(psb->sd >= 0 && psb->sd < 4096) {
      if(FD_ISSET(psb->sd, prd)) {
        psb->handler(psb->sd);
        return;
      }
    } else {
      printf("something got corrupted\n");
      exit(EXIT_FAILURE);
    }
  }
}

/* vim:set ts=2: */
/*
 * Local Variables:
 * tab-width: 2
 * c-indent-level: 2
 * c-basic-offset: 2
 * End:
 */
